﻿Imports System.Collections.Generic
Imports System.Text
Imports System.Runtime.InteropServices
Imports System.IO
'This class was found at the following URL:
'http://heskandari.blogspot.com/2008/03/structured-storage.html
Namespace StructuredStorage
#Region "Structured Storage Exception"

    ''' <summary>
    ''' Exception thrown by Structure Storage.
    ''' </summary>
    Public Class StructuredStorageException
        Inherits ApplicationException
        Public Sub New(ByVal message As String)
            MyBase.New(message)
        End Sub

        Public Sub New(ByVal message As String, ByVal inner As Exception)
            MyBase.New(message, inner)
        End Sub

        Public Sub New()
        End Sub
    End Class

#End Region

#Region "Structured Storage Wrapper"

    ''' <summary>
    ''' Provides native access to Structured Storage of
    ''' the NTFS file system.
    ''' </summary>
    Public Class StructuredStorage
        Implements IDisposable
#Region "Interface Ids"

        Private Shared PropertySetStorageId As New Guid("0000013A-0000-0000-C000-000000000046")
        Private Shared SummaryPropertiesId As New Guid("F29F85E0-4FF9-1068-AB91-08002B27B3D9")
        Private propSetStorage As IPropertySetStorage = Nothing
        Private propStorage As IPropertyStorage = Nothing

#End Region

#Region "Fields"

        Private fi As FileInfo

        Public Const WordExt As String = ".doc"
        Public Const ExcelExt As String = ".xls"
        Public Const PowerPointExt As String = ".ppt"

#End Region

#Region "Ctor"

        Public Sub New(ByVal fi As FileInfo)
            Me.fi = fi
            If Not fi.Exists Then
                Throw New StructuredStorageException(String.Format("specified file was not found : {0}", fi.FullName))
            End If
        End Sub

        Public Sub New(ByVal filePath As String)
            Me.fi = New FileInfo(filePath)
            If Not fi.Exists Then
                Throw New StructuredStorageException(String.Format("specified file was not found : {0}", filePath))
            End If
        End Sub

#End Region

#Region "Methods"

        ''' <summary>
        ''' Updates the provided property fo a file.
        ''' </summary>
        ''' <param name="fi"></param>
        ''' <param name="property"></param>
        ''' <param name="value"></param>
        Public Sub WriteProperty(ByVal [property] As SummaryProperty, ByVal value As Object)
            If Not ExtendedPropertiesSupported Then
                Return
            End If

            Try
                OpenStorage(fi.FullName, CInt(AccessMode.ShareExclusive Or AccessMode.ReadWrite), CInt(FileFormat.Any), 0, CType(0, IntPtr), IntPtr.Zero, _
                 PropertySetStorageId, propSetStorage)
                propSetStorage.Create(SummaryPropertiesId, IntPtr.Zero, CInt(PropertySetFlags.[Default]), CInt(AccessMode.Create Or AccessMode.ReadWrite Or AccessMode.ShareExclusive), propStorage)

                ' next, you assemble a property descriptor for the property you 
                ' want to write to, in our case the Comment property 
                Dim propertySpecification As New PropSpec()

                propertySpecification.ulKind = CInt(PropSpecTypes.PRSPEC_PROPID)
                propertySpecification.Name_Or_ID = New IntPtr(CInt([property]))

                'now, set the value you want in a property variant 
                Dim propertyValue As New PropVariant()
                propertyValue.FromObject(value)

                ' Simply pass the property spec and its new value to the WriteMultiple 
                ' method and you're almost done 
                propStorage.WriteMultiple(1, propertySpecification, propertyValue, 2)

                ' the only thing left to do is commit your changes. Now you're done! 
                propStorage.Commit(CInt(CommitFlags.[Default]))
            Catch ex As Exception
                Throw New StructuredStorageException("There was an error accessing the structured storage for file " & fi.FullName, ex)
            End Try
        End Sub

        Public Function ReadProperty(ByVal [property] As SummaryProperty) As Object
            If Not ExtendedPropertiesSupported Then
                Return Nothing
            End If

            Dim result As String = String.Empty

            Try
                OpenStorage(fi.FullName, CInt(AccessMode.Read Or AccessMode.ShareExclusive), CInt(FileFormat.Any), 0, IntPtr.Zero, IntPtr.Zero, _
                 PropertySetStorageId, propSetStorage)
                propSetStorage.Open(SummaryPropertiesId, CInt(AccessMode.Read Or AccessMode.ShareExclusive), propStorage)

                Dim spec As New PropSpec()
                spec.Name_Or_ID = New IntPtr(CInt([property]))
                spec.ulKind = CInt(PropSpecTypes.PRSPEC_PROPID)
                Dim propertyValue As New PropVariant()
                propStorage.ReadMultiple(1, spec, propertyValue)
                result = Marshal.PtrToStringAnsi(propertyValue.pointerValue)
            Catch ex As Exception
                Throw New StructuredStorageException("There was an error accessing the structured storage for file " & fi.FullName, ex)
            End Try

            Return result
        End Function

        ''' <summary>
        ''' Updates the comment of a file
        ''' </summary>
        ''' <param name="fi"></param>
        ''' <param name="comment"></param>
        Public Sub UpdateComment(ByVal comment As String)
            WriteProperty(SummaryProperty.Comments, comment)
        End Sub

        Private ReadOnly Property ExtendedPropertiesSupported() As Boolean
            Get
                If fi Is Nothing OrElse Not fi.Exists Then
                    Return False
                End If

                Dim di As New DirectoryInfo(fi.FullName)
                Dim drive As New DriveInfo(di.Root.Name)

                Return drive.DriveFormat.ToUpper() = "NTFS" OrElse IsOfficeExtension(fi.Extension)
            End Get
        End Property

        Private Function IsOfficeExtension(ByVal extension As String) As Boolean
            If String.IsNullOrEmpty(extension) Then
                Return False
            End If

            Return extension.Equals(WordExt, StringComparison.OrdinalIgnoreCase) OrElse extension.Equals(ExcelExt, StringComparison.OrdinalIgnoreCase) OrElse extension.Equals(PowerPointExt, StringComparison.OrdinalIgnoreCase)
        End Function

#End Region

#Region "Wrapper"

#Region "IDisposable"

        Public Sub Dispose() Implements IDisposable.Dispose
            If propSetStorage IsNot Nothing Then
                Marshal.ReleaseComObject(propSetStorage)
                propSetStorage = Nothing
            End If

            If propStorage IsNot Nothing Then
                Marshal.ReleaseComObject(propStorage)
                propStorage = Nothing
            End If
        End Sub

#End Region

#Region "PropVariant"

        <StructLayout(LayoutKind.Explicit, Size:=16)> _
        Private Structure PropVariant
            <FieldOffset(0)> _
            Public variantType As Short
            <FieldOffset(8)> _
            Public pointerValue As IntPtr
            <FieldOffset(8)> _
            Public byteValue As Byte
            <FieldOffset(8)> _
            Public longValue As Long


            Public Sub FromObject(ByVal obj As Object)
                If obj.[GetType]() Is GetType(String) Then
                    Me.variantType = CType(System.Runtime.InteropServices.VarEnum.VT_LPWSTR, Short)
                    Me.pointerValue = Marshal.StringToHGlobalUni(DirectCast(obj, String))
                End If
            End Sub
        End Structure

#End Region

#Region "IPropertySetStorage"

        <ComVisible(True)> _
        <ComImport()> _
        <Guid("0000013A-0000-0000-C000-000000000046")> _
        <InterfaceType(ComInterfaceType.InterfaceIsIUnknown)> _
        Private Interface IPropertySetStorage
            Function Create(<[In](), MarshalAs(UnmanagedType.Struct)> ByRef rfmtid As System.Guid, <[In]()> ByVal pclsid As IntPtr, <[In]()> ByVal grfFlags As Integer, <[In]()> ByVal grfMode As Integer, ByRef propertyStorage As IPropertyStorage) As UInteger


            Function Open(<[In](), MarshalAs(UnmanagedType.Struct)> ByRef rfmtid As System.Guid, <[In]()> ByVal grfMode As Integer, ByRef propertyStorage As IPropertyStorage) As Integer
        End Interface

#End Region

#Region "PropSpec"

        <StructLayout(LayoutKind.Explicit, Size:=8, CharSet:=CharSet.Unicode)> _
        Private Structure PropSpec
            <FieldOffset(0)> _
            Public ulKind As Integer
            <FieldOffset(4)> _
            Public Name_Or_ID As IntPtr
        End Structure

#End Region

        ''' <summary>
        ''' Property ID types
        ''' </summary>
        Private Enum PropSpecTypes As UInteger
            PRSPEC_LPWSTR = 0
            PRSPEC_PROPID = 1
            PRSPEC_INVALID = &HFFFFFFFFUI
        End Enum

#Region "IPropertyStorage"

        <ComVisible(True)> _
        <ComImport()> _
        <Guid("00000138-0000-0000-C000-000000000046")> _
        <InterfaceType(ComInterfaceType.InterfaceIsIUnknown)> _
        Private Interface IPropertyStorage
            Function ReadMultiple(ByVal numProperties As UInteger, <MarshalAs(UnmanagedType.Struct)> ByRef propertySpecifications As PropSpec, <MarshalAs(UnmanagedType.Struct)> ByRef propertyValues As PropVariant) As Integer

            Function WriteMultiple(ByVal numProperties As UInteger, <MarshalAs(UnmanagedType.Struct)> ByRef propertySpecification As PropSpec, ByRef propertyValues As PropVariant, ByVal propIDNameFirst As Integer) As Integer


            Function Commit(ByVal commitFlags As Integer) As UInteger
        End Interface

#End Region

#Region "PropertySetFlags"

        Private Enum PropertySetFlags
            [Default] = 0
            NonSimple = 1
            Ansi = 2
            UnBuffered = 4
            CaseSensitive = 8
        End Enum

#End Region

#Region "HResults"

        Private Enum HResults As UInteger
            S_OK = 0
            STG_E_FILEALREADYEXISTS = &H80030050UI
        End Enum

#End Region

#Region "STGOptions"

        <StructLayout(LayoutKind.Explicit, Size:=12, CharSet:=CharSet.Unicode)> _
        Private Structure STGOptions
            <FieldOffset(0)> _
            Private usVersion As UShort
            <FieldOffset(2)> _
            Private reserved As UShort
            <FieldOffset(4)> _
            Private uiSectorSize As UInteger
            <FieldOffset(8), MarshalAs(UnmanagedType.LPWStr)> _
            Private pwcsTemplateFile As String
        End Structure

#End Region

#Region "Storage Commit Flags"

        Public Enum CommitFlags
            [Default] = 0
            Overwrite = 1
            OnlyIfCurrent = 2
            DangerouslyCommitMerelyToDiskCache = 4
            Consolidate = 8
        End Enum

#End Region

#End Region

#Region "Win32"

        <DllImport("Ole32.dll", EntryPoint:="StgOpenStorageEx", CharSet:=CharSet.Unicode)> _
        Private Shared Function OpenStorage(<MarshalAs(UnmanagedType.LPWStr)> ByVal name As String, ByVal accessMode As Integer, ByVal storageFileFormat As Integer, ByVal fileBuffering As Integer, ByVal options As IntPtr, ByVal reserved As IntPtr, _
         ByRef riid As System.Guid, <MarshalAs(UnmanagedType.[Interface])> ByRef propertySetStorage As IPropertySetStorage) As UInteger
        End Function

#End Region
    End Class

#End Region

#Region "Summary Property"

    Public Enum SummaryProperty
        Title = &H2
        Subject = &H3
        Author = &H4
        Keywords = &H5
        Comments = &H6
        Template = &H7
        LastSavedBy = &H8
        RevisionNumber = &H9
        TotalEditingTime = &HA
        LastPrinted = &HB
        CreateDateTime = &HC
        LastSaveDateTime = &HD
        NumberOfPages = &HE
        NumberOfWords = &HF
        NumberOfChars = &H10
        Thumbnail = &H11
        AppName = &H12
        Security = &H13
    End Enum

#End Region

#Region "Storage Access Mode"

    Public Enum AccessMode
        Read = &H0
        Write = &H1
        ReadWrite = &H2
        ShareDenyNone = &H40
        ShareDenyRead = &H30
        ShareDenyWrite = &H20
        ShareExclusive = &H10
        Priority = &H40000
        Create = &H1000
        Convert = &H20000
        FailIfThere = &H0
        Direct = &H0
        Transacted = &H10000
        NoScratch = &H100000
        NoSnapshot = &H200000
        Simple = &H8000000
        DirectSwmr = &H400000
        DeleteOnRelease = &H4000000
    End Enum

#End Region

#Region "Storage File Format"

    Public Enum FileFormat
        Storage = 0
        File = 3
        Any = 4
        DocFile = 5
    End Enum

#End Region
End Namespace

Public Class clsStorage

End Class
